@@ -1,10 +1,12 @@
-/* $Id: minissdpd.c,v 1.18 2008/10/07 12:24:54 nanard Exp $ */
+/* $Id: minissdpd.c,v 1.22 2011/07/29 08:44:04 nanard Exp $ */
 /* MiniUPnP project
- * (c) 2007-2008 Thomas Bernard
+ * (c) 2007-2011 Thomas Bernard
  * website : http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -24,8 +26,11 @@
 #include <sys/stat.h>
 /* unix sockets */
 #include <sys/un.h>
+/* for getpwnam() and getgrnam() */
+#include <pwd.h>
+#include <grp.h>
 
-/*#include "minissdp.h"*/
+#include "upnputils.h"
 #include "openssdpsocket.h"
 #include "daemonize.h"
 #include "codelength.h"
@@ -59,9 +64,14 @@ struct device {
 /* discovered device list kept in memory */
 struct device * devlist = 0;
 
+/* bootid and configid */
+unsigned int upnp_bootid = 1;
+unsigned int upnp_configid = 1337;
+
 /* updateDevice() :
  * adds or updates the device to the list.
  * return value :
+ *  -1 : error
  *   0 : the device was updated
  *   1 : the device was new    */
 int updateDevice(const struct header * headers, time_t t)
@@ -102,6 +112,10 @@ int updateDevice(const struct header * headers, time_t t)
 		int i;
 		p = malloc(  sizeof(struct device)
 		           + headers[0].l+headers[1].l+headers[2].l );
+		if(!p) {
+			syslog(LOG_ERR, "updateDevice(): cannot allocate memory");
+			return -1;
+		}
 		p->next = devlist;
 		p->t = t;
 		pc = p->data;
@@ -148,35 +162,49 @@ int removeDevice(const struct header * headers)
 /* SendSSDPMSEARCHResponse() :
  * build and send response to M-SEARCH SSDP packets. */
 static void
-SendSSDPMSEARCHResponse(int s, const struct sockaddr_in * sockname,
+SendSSDPMSEARCHResponse(int s, const struct sockaddr * sockname,
                         const char * st, const char * usn,
                         const char * server, const char * location)
 {
-    int l, n;
-    char buf[512];
-    /*
-     * follow guideline from document "UPnP Device Architecture 1.0"
-     * uppercase is recommended.
-     * DATE: is recommended
-     * SERVER: OS/ver UPnP/1.0 miniupnpd/1.0
-     * - check what to put in the 'Cache-Control' header
-     * */
-    l = snprintf(buf, sizeof(buf), "HTTP/1.1 200 OK\r\n"
-        "CACHE-CONTROL: max-age=120\r\n"
-        /*"DATE: ...\r\n"*/
-        "ST: %s\r\n"
-        "USN: %s\r\n"
-        "EXT:\r\n"
-        "SERVER: %s\r\n"
-        "LOCATION: %s\r\n"
-        "\r\n",
-        st, usn,
-        server, location);
-    n = sendto(s, buf, l, 0,
-               (struct sockaddr *)sockname, sizeof(struct sockaddr_in) );
-    if(n < 0) {
-        syslog(LOG_ERR, "sendto(udp): %m");
-    }
+	int l, n;
+	char buf[512];
+	socklen_t sockname_len;
+	/*
+	 * follow guideline from document "UPnP Device Architecture 1.0"
+	 * uppercase is recommended.
+	 * DATE: is recommended
+	 * SERVER: OS/ver UPnP/1.0 miniupnpd/1.0
+	 * - check what to put in the 'Cache-Control' header
+	 *
+	 * have a look at the document "UPnP Device Architecture v1.1 */
+	l = snprintf(buf, sizeof(buf), "HTTP/1.1 200 OK\r\n"
+		"CACHE-CONTROL: max-age=120\r\n"
+		/*"DATE: ...\r\n"*/
+	"ST: %s\r\n"
+	"USN: %s\r\n"
+	"EXT:\r\n"
+	"SERVER: %s\r\n"
+	"LOCATION: %s\r\n"
+		"OPT: \"http://schemas.upnp.org/upnp/1/0/\";\r\n" /* UDA v1.1 */
+		"01-NLS: %u\r\n" /* same as BOOTID. UDA v1.1 */
+		"BOOTID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
+		"CONFIGID.UPNP.ORG: %u\r\n" /* UDA v1.1 */
+	"\r\n",
+		st, usn,
+		server, location,
+		upnp_bootid, upnp_bootid, upnp_configid);
+#ifdef ENABLE_IPV6
+	sockname_len = (sockname->sa_family == PF_INET6)
+	             ? sizeof(struct sockaddr_in6)
+	             : sizeof(struct sockaddr_in);
+#else
+	sockname_len = sizeof(struct sockaddr_in);
+#endif
+	n = sendto(s, buf, l, 0,
+	           sockname, sockname_len );
+	if(n < 0) {
+		syslog(LOG_ERR, "sendto(udp): %m");
+	}
 }
 
 /* Services stored for answering to M-SEARCH */
@@ -191,15 +219,25 @@ LIST_HEAD(servicehead, service) servicelisthead;
 
 /* Process M-SEARCH requests */
 void processMSEARCH(int s, const char * st, int st_len,
-                    const struct sockaddr_in * addr)
+                    const struct sockaddr * addr)
 {
 	struct service * serv;
+#ifdef ENABLE_IPV6
+	char buf[64];
+#endif
+
 	if(!st || st_len==0)
 		return;
+#ifdef ENABLE_IPV6
+	sockaddr_to_string(addr, buf, sizeof(buf));
+	syslog(LOG_INFO, "SSDP M-SEARCH from %s ST:%.*s",
+	       buf, st_len, st);
+#else
 	syslog(LOG_INFO, "SSDP M-SEARCH from %s:%d ST: %.*s",
-           inet_ntoa(addr->sin_addr),
-           ntohs(addr->sin_port),
-           st_len, st);
+	       inet_ntoa(((const struct sockaddr_in *)addr)->sin_addr),
+	       ntohs(((const struct sockaddr_in *)addr)->sin_port),
+	       st_len, st);
+#endif
 	if(st_len==8 && (0==memcmp(st, "ssdp:all", 8))) {
 		/* send a response for all services */
 		for(serv = servicelisthead.lh_first;
@@ -247,7 +285,7 @@ void processMSEARCH(int s, const char * st, int st_len,
  *     0 : no device removed nor added
  *     1 : a device was added.  */
 int ParseSSDPPacket(int s, const char * p, ssize_t n,
-                    const struct sockaddr_in * addr)
+                    const struct sockaddr * addr)
 {
 	const char * linestart;
 	const char * lineend;
@@ -273,28 +311,32 @@ int ParseSSDPPacket(int s, const char * p, ssize_t n,
 	while(linestart < p + n - 2) {
 		/* start parsing the line : detect line end */
 		lineend = linestart;
-		while(*lineend != '\n' && *lineend != '\r' && lineend < p + n)
+		while(lineend < p + n && *lineend != '\n' && *lineend != '\r')
 			lineend++;
 		//printf("line: '%.*s'\n", lineend - linestart, linestart);
 		/* detect name end : ':' character */
 		nameend = linestart;
-		while(*nameend != ':' && nameend < lineend)
+		while(nameend < lineend && *nameend != ':')
 			nameend++;
 		/* detect value */
-		valuestart = nameend + 1;
-		while(isspace(*valuestart) && valuestart < lineend)
+		if(nameend < lineend)
+			valuestart = nameend + 1;
+		else
+			valuestart = nameend;
+		/* trim spaces */
+		while(valuestart < lineend && isspace(*valuestart))
 			valuestart++;
 		/* suppress leading " if needed */
-		if(*valuestart=='\"')
+		if(valuestart < lineend && *valuestart=='\"')
 			valuestart++;
 		if(nameend > linestart && valuestart < lineend) {
-			int l = nameend - linestart;
-			int m = lineend - valuestart;
+			int l = nameend - linestart;	/* header name length */
+			int m = lineend - valuestart;	/* header value length */
 			/* suppress tailing spaces */
 			while(m>0 && isspace(valuestart[m-1]))
 				m--;
 			/* suppress tailing ' if needed */
-			if(valuestart[m-1] == '\"')
+			if(m>0 && valuestart[m-1] == '\"')
 				m--;
 			i = -1;
 			/*printf("--%.*s: (%d)%.*s--\n", l, linestart,
@@ -312,22 +354,29 @@ int ParseSSDPPacket(int s, const char * p, ssize_t n,
 			else if(l==8 && 0==strncasecmp(linestart, "location", 8))
 				i = HEADER_LOCATION;
 			else if(l==13 && 0==strncasecmp(linestart, "cache-control", 13)) {
-				const char * name = valuestart;
-				const char * val;
-				int rem = m;
+				/* parse "name1=value1, name_alone, name2=value2" string */
+				const char * name = valuestart;	/* name */
+				const char * val;				/* value */
+				int rem = m;	/* remaining bytes to process */
 				while(rem > 0) {
 					val = name;
-					while(*val != '=' && *val != ',')
+					while(val < name + rem && *val != '=' && *val != ',')
 						val++;
+					if(val >= name + rem)
+						break;
 					if(*val == '=') {
-						while(*val == '=' || isspace(*val))
+						while(val < name + rem && (*val == '=' || isspace(*val)))
 							val++;
+						if(val >= name + rem)
+							break;
 						if(0==strncasecmp(name, "max-age", 7))
 							lifetime = (unsigned int)strtoul(val, 0, 0);
+						/* move to the next name=value pair */
 						while(rem > 0 && *name != ',') {
 							rem--;
 							name++;
 						}
+						/* skip spaces */
 						while(rem > 0 && (*name == ',' || isspace(*name))) {
 							rem--;
 							name++;
@@ -435,26 +484,26 @@ void processRequest(struct reqelem * req)
 	unsigned char * rp = rbuf+1;
 	unsigned char nrep = 0;
 	time_t t;
-	struct service * newserv;
+	struct service * newserv = NULL;
 	struct service * serv;
 
 	n = read(req->socket, buf, sizeof(buf));
 	if(n<0) {
 		syslog(LOG_ERR, "(s=%d) processRequest(): read(): %m", req->socket);
-		close(req->socket);
-		req->socket = -1;
-		return;
+		goto error;
 	}
 	if(n==0) {
 		syslog(LOG_INFO, "(s=%d) request connection closed", req->socket);
-		close(req->socket);
-		req->socket = -1;
-		return;
+		goto error;
 	}
 	t = time(NULL);
 	type = buf[0];
 	p = buf + 1;
 	DECODELENGTH(l, p);
+	if(p+l > buf+n) {
+		syslog(LOG_WARNING, "bad request (length encoding)");
+		goto error;
+	}
 	syslog(LOG_INFO, "(s=%d) request type=%d str='%.*s'",
 	       req->socket, type, l, p);
 	switch(type) {
@@ -497,6 +546,7 @@ void processRequest(struct reqelem * req)
 		for(serv = servicelisthead.lh_first;
 		    serv && (nrep < 255);
 		    serv = serv->entries.le_next) {
+			/* test if we can put more responses in the buffer */
 			if(strlen(serv->location) + strlen(serv->st)
 			  + strlen(serv->usn) + 6 + (rp - rbuf) >= sizeof(rbuf))
 			  	break;
@@ -523,29 +573,39 @@ void processRequest(struct reqelem * req)
 			}
 		}
 		rbuf[0] = nrep;
-		if(write(req->socket, rbuf, rp - rbuf) < 0)
+		if(write(req->socket, rbuf, rp - rbuf) < 0) {
 			syslog(LOG_ERR, "(s=%d) write: %m", req->socket);
+			goto error;
+		}
 		break;
 	case 4:
 		newserv = malloc(sizeof(struct service));
 		if(!newserv) {
 			syslog(LOG_ERR, "cannot allocate memory");
-			return;
+			goto error;
 		}
 		newserv->st = malloc(l + 1);
 		if(!newserv->st) {
 			syslog(LOG_ERR, "cannot allocate memory");
-			return;
+			goto error;
 		}
 		memcpy(newserv->st, p, l);
 		newserv->st[l] = '\0';
 		p += l;
+		if(p >= buf + n) {
+			syslog(LOG_WARNING, "bad request (missing usn)");
+			goto error;
+		}
 		DECODELENGTH(l, p);
+		if(p+l > buf+n) {
+			syslog(LOG_WARNING, "bad request (length encoding)");
+			goto error;
+		}
 		syslog(LOG_INFO, "usn='%.*s'", l, p);
 		newserv->usn = malloc(l + 1);
 		if(!newserv->usn) {
 			syslog(LOG_ERR, "cannot allocate memory");
-			return;
+			goto error;
 		}
 		memcpy(newserv->usn, p, l);
 		newserv->usn[l] = '\0';
@@ -555,7 +615,7 @@ void processRequest(struct reqelem * req)
 		newserv->server = malloc(l + 1);
 		if(!newserv->server) {
 			syslog(LOG_ERR, "cannot allocate memory");
-			return;
+			goto error;
 		}
 		memcpy(newserv->server, p, l);
 		newserv->server[l] = '\0';
@@ -565,7 +625,7 @@ void processRequest(struct reqelem * req)
 		newserv->location = malloc(l + 1);
 		if(!newserv->location) {
 			syslog(LOG_ERR, "cannot allocate memory");
-			return;
+			goto error;
 		}
 		memcpy(newserv->location, p, l);
 		newserv->location[l] = '\0';
@@ -583,11 +643,13 @@ void processRequest(struct reqelem * req)
 				free(serv->location);
 				serv->location = newserv->location;
 				free(newserv);
+				newserv = NULL;
 				return;
 			}
 		}
 		/* Inserting new service */
 		LIST_INSERT_HEAD(&servicelisthead, newserv, entries);
+		newserv = NULL;
 		/*rbuf[0] = '\0';
 		if(write(req->socket, rbuf, 1) < 0)
 			syslog(LOG_ERR, "(s=%d) write: %m", req->socket);
@@ -596,9 +658,24 @@ void processRequest(struct reqelem * req)
 	default:
 		syslog(LOG_WARNING, "Unknown request type %d", type);
 		rbuf[0] = '\0';
-		if(write(req->socket, rbuf, 1) < 0)
+		if(write(req->socket, rbuf, 1) < 0) {
 			syslog(LOG_ERR, "(s=%d) write: %m", req->socket);
+			goto error;
+		}
 	}
+	return;
+error:
+	if(newserv) {
+		free(newserv->st);
+		free(newserv->usn);
+		free(newserv->server);
+		free(newserv->location);
+		free(newserv);
+		newserv = NULL;
+	}
+	close(req->socket);
+	req->socket = -1;
+	return;
 }
 
 static volatile int quitting = 0;
@@ -614,12 +691,16 @@ sigterm(int sig)
 /* main(): program entry point */
 int main(int argc, char * * argv)
 {
+	int ret = 0;
 	int pid;
 	struct sigaction sa;
 	char buf[1500];
 	ssize_t n;
-	int s_ssdp;	/* udp socket receiving ssdp packets */
-	int s_unix;	/* unix socket communicating with clients */
+	int s_ssdp = -1;	/* udp socket receiving ssdp packets */
+#ifdef ENABLE_IPV6
+	int s_ssdp6 = -1;	/* udp socket receiving ssdp packets IPv6*/
+#endif
+	int s_unix = -1;	/* unix socket communicating with clients */
 	int s;
 	LIST_HEAD(reqstructhead, reqelem) reqlisthead;
 	struct reqelem * req;
@@ -631,10 +712,14 @@ int main(int argc, char * * argv)
 	const char * sockpath = "/var/run/minissdpd.sock";
 	const char * pidfilename = "/var/run/minissdpd.pid";
 	int debug_flag = 0;
+	int ipv6 = 0;
 	int deltadev = 0;
 	struct sockaddr_in sendername;
-	socklen_t len_r;
-	len_r = sizeof(struct sockaddr_in);
+	socklen_t sendername_len;
+#ifdef ENABLE_IPV6
+	struct sockaddr_in6 sendername6;
+	socklen_t sendername6_len;
+#endif
 
 	LIST_INIT(&reqlisthead);
 	LIST_INIT(&servicelisthead);
@@ -653,15 +738,21 @@ int main(int argc, char * * argv)
 			sockpath = argv[++i];
 		else if(0==strcmp(argv[i], "-p"))
 			pidfilename = argv[++i];
+#ifdef ENABLE_IPV6
+		else if(0==strcmp(argv[i], "-6"))
+			ipv6 = 1;
+#endif
 	}
 	if(n_if_addr < 1)
 	{
 		fprintf(stderr,
-		        "Usage: %s [-d] [-s socket] [-p pidfile] "
-				"-i <interface_address> [-i <interface_address2>] ...\n",
+		        "Usage: %s [-d] [-6] [-s socket] [-p pidfile] "
+				"-i <interface> [-i <interface2>] ...\n",
 		        argv[0]);
 		fprintf(stderr,
-		        "By default, socket will be open as %s "
+		        "\n  <interface> is either an IPv4 address such as 192.168.1.42, or an\ninterface name such as eth0.\n");
+		fprintf(stderr,
+		        "\n  By default, socket will be open as %s\n"
 		        "and pid written to file %s\n",
 		        sockpath, pidfilename);
 		return 1;
@@ -700,32 +791,86 @@ int main(int argc, char * * argv)
 	if(sigaction(SIGTERM, &sa, NULL))
 	{
 		syslog(LOG_ERR, "Failed to set SIGTERM handler. EXITING");
-		return 1;
+		ret = 1;
+		goto quit;
 	}
 	if(sigaction(SIGINT, &sa, NULL))
 	{
 		syslog(LOG_ERR, "Failed to set SIGINT handler. EXITING");
-		return 1;
+		ret = 1;
+		goto quit;
 	}
-	s_ssdp = OpenAndConfSSDPReceiveSocket(n_if_addr, if_addr);
+	/* open UDP socket(s) for receiving SSDP packets */
+	s_ssdp = OpenAndConfSSDPReceiveSocket(n_if_addr, if_addr, 0);
 	if(s_ssdp < 0)
 	{
 		syslog(LOG_ERR, "Cannot open socket for receiving SSDP messages, exiting");
-		return 1;
+		ret = 1;
+		goto quit;
 	}
+#ifdef ENABLE_IPV6
+	if(ipv6) {
+		s_ssdp6 = OpenAndConfSSDPReceiveSocket(n_if_addr, if_addr, 1);
+		if(s_ssdp6 < 0)
+		{
+			syslog(LOG_ERR, "Cannot open socket for receiving SSDP messages (IPv6), exiting");
+			ret = 1;
+			goto quit;
+		}
+	}
+#endif
+	/* Open Unix socket to communicate with other programs on
+	 * the same machine */
 	s_unix = OpenUnixSocket(sockpath);
 	if(s_unix < 0)
 	{
 		syslog(LOG_ERR, "Cannot open unix socket for communicating with clients. Exiting");
-		return 1;
+		ret = 1;
+		goto quit;
 	}
 
+	/* drop privileges */
+#if 0
+	/* if we drop privileges, how to unlink(/var/run/minissdpd.sock) ? */
+	if(getuid() == 0) {
+		struct passwd * user;
+		struct group * group;
+		user = getpwnam("nobody");
+		if(!user) {
+			syslog(LOG_ERR, "getpwnam(\"%s\") : %m", "nobody");
+			ret = 1;
+			goto quit;
+		}
+		group = getgrnam("nogroup");
+		if(!group) {
+			syslog(LOG_ERR, "getgrnam(\"%s\") : %m", "nogroup");
+			ret = 1;
+			goto quit;
+		}
+		if(setgid(group->gr_gid) < 0) {
+			syslog(LOG_ERR, "setgit(%d) : %m", group->gr_gid);
+			ret = 1;
+			goto quit;
+		}
+		if(setuid(user->pw_uid) < 0) {
+			syslog(LOG_ERR, "setuid(%d) : %m", user->pw_uid);
+			ret = 1;
+			goto quit;
+		}
+	}
+#endif
+
 	/* Main loop */
 	while(!quitting)
 	{
 		/* fill readfds fd_set */
 		FD_ZERO(&readfds);
 		FD_SET(s_ssdp, &readfds);
+#ifdef ENABLE_IPV6
+		if(s_ssdp6 >= 0) {
+			FD_SET(s_ssdp6, &readfds);
+		}
+#endif
 		FD_SET(s_unix, &readfds);
 		for(req = reqlisthead.lh_first; req; req = req->entries.le_next)
 		{
@@ -739,10 +884,12 @@ int main(int argc, char * * argv)
 				syslog(LOG_ERR, "select: %m");
 			break;
 		}
-		if(FD_ISSET(s_ssdp, &readfds))
+#ifdef ENABLE_IPV6
+		if((s_ssdp6 >= 0) && FD_ISSET(s_ssdp6, &readfds))
 		{
-			n = recvfrom(s_ssdp, buf, sizeof(buf), 0,
-			             (struct sockaddr *)&sendername, &len_r);
+			sendername6_len = sizeof(struct sockaddr_in6);
+			n = recvfrom(s_ssdp6, buf, sizeof(buf), 0,
+			             (struct sockaddr *)&sendername6, &sendername6_len);
 			if(n<0)
 			{
 				syslog(LOG_ERR, "recvfrom: %m");
@@ -751,7 +898,8 @@ int main(int argc, char * * argv)
 			{
 				/* Parse and process the packet received */
 				/*printf("%.*s", n, buf);*/
-				i = ParseSSDPPacket(s_ssdp, buf, n, &sendername);
+				i = ParseSSDPPacket(s_ssdp6, buf, n,
+				                    (struct sockaddr *)&sendername6);
 				syslog(LOG_DEBUG, "** i=%d deltadev=%d **", i, deltadev);
 				if(i==0 || (i*deltadev < 0))
 				{
@@ -767,6 +915,38 @@ int main(int argc, char * * argv)
 				}
 			}
 		}
+#endif
+		if(FD_ISSET(s_ssdp, &readfds))
+		{
+			sendername_len = sizeof(struct sockaddr_in);
+			n = recvfrom(s_ssdp, buf, sizeof(buf), 0,
+			             (struct sockaddr *)&sendername, &sendername_len);
+			if(n<0)
+			{
+				syslog(LOG_ERR, "recvfrom: %m");
+			}
+			else
+			{
+				/* Parse and process the packet received */
+				/*printf("%.*s", n, buf);*/
+				i = ParseSSDPPacket(s_ssdp, buf, n,
+				                    (struct sockaddr *)&sendername);
+				syslog(LOG_DEBUG, "** i=%d deltadev=%d **", i, deltadev);
+				if(i==0 || (i*deltadev < 0))
+				{
+					if(deltadev > 0)
+						syslog(LOG_NOTICE, "%d new devices added", deltadev);
+					else if(deltadev < 0)
+						syslog(LOG_NOTICE, "%d devices removed (good-bye!)", -deltadev);
+					deltadev = i;
+				}
+				else if((i*deltadev) >= 0)
+				{
+					deltadev += i;
+				}
+			}
+		}
+		/* processing unix socket requests */
 		for(req = reqlisthead.lh_first; req;)
 		{
 			reqnext = req->entries.le_next;
@@ -781,6 +961,7 @@ int main(int argc, char * * argv)
 			}
 			req = reqnext;
 		}
+		/* processing new requests */
 		if(FD_ISSET(s_unix, &readfds))
 		{
 			struct reqelem * tmp;
@@ -793,19 +974,41 @@ int main(int argc, char * * argv)
 			{
 				syslog(LOG_INFO, "(s=%d) new request connection", s);
 				tmp = malloc(sizeof(struct reqelem));
-				tmp->socket = s;
-				LIST_INSERT_HEAD(&reqlisthead, tmp, entries);
+				if(!tmp)
+				{
+					syslog(LOG_ERR, "cannot allocate memory for request");
+					close(s);
+				}
+				else
+				{
+					tmp->socket = s;
+					LIST_INSERT_HEAD(&reqlisthead, tmp, entries);
+				}
 			}
 		}
 	}
+
 	/* closing and cleaning everything */
-	close(s_ssdp);
-	close(s_unix);
+quit:
+	if(s_ssdp >= 0) {
+		close(s_ssdp);
+		s_ssdp = -1;
+	}
+#ifdef ENABLE_IPV6
+	if(s_ssdp6 >= 0) {
+		close(s_ssdp6);
+		s_ssdp6 = -1;
+	}
+#endif
+	if(s_unix >= 0) {
+		close(s_unix);
+		s_unix = -1;
+	}
 	if(unlink(sockpath) < 0)
 		syslog(LOG_ERR, "unlink(%s): %m", sockpath);
 	if(unlink(pidfilename) < 0)
 		syslog(LOG_ERR, "unlink(%s): %m", pidfilename);
 	closelog();
-	return 0;
+	return ret;
 }
 
